﻿// Dependencies
//#region Dependencies

// External (Common) dependencies
var fs = require('fs');
var cp = require('child_process');
var moment = require('moment');

// External (Device) dependencies
var Parser = require('expr-eval').Parser;// Library to manage parsing and evaluation of formulas
var mathparser = new Parser();

// Local dependencies
var Logger;// Used for logging
var SocketIo;// Used for messaging
var NetSnmp;// Used for SNMP messaging - base library
var SmtpManager;// Used for SMTP messaging
var SnmpHostSessionsArray = [];// Instance array for list of SNMP host sessions

//#endregion Dependencies

// Constants
//#region Constants

// Constants (Common)
//#region Constants (Common)
//  Device
var DEVICE_MODEL = "./NonDirectionalPower.json"; //  Device model defines the device incl. general properties, parameters and I/O 

//  Event Log
//#region Event Log
var EVENTS_LENGTH = 10;
var ThresholdValues = ["NORMAL", "BELOW", "ABOVE"];
//#endregion Event Log
//#endregion Constants (Common)

// Constants (Device)
//#region Constants (Device)
// Holding States
var HOLDSTATE_OFF = "red";// Active
var HOLDSTATE_ACTIVE = "green";// Active
var HOLDSTATE_HOLDING = "#8F8F8F"// Changing
var HOLDSTATE_UNKNOWN = "#EB6F21";// Orangy

// Analog Inputs
//#region Analog Inputs
//  Analog input offset will be determined by selected port #
var ANALOG_INPUTS_LOCATION = '/sys/class/hwmon/hwmon0';

//  Analog input raw to voltage linear equation
var ANALOG_INPUTS_RAWVALUE_OFFSET = -0.075;
var ANALOG_INPUTS_RAWVALUE_SLOPE = 0.005;
var ANALOG_INPUTS_CHANGE_FILTER = 4;// Look for 20mV change
var ANALOG_INPUTS_TTL_ONOFF = 2.00;// Amount of voltage to switch to ON
var ANALOG_INPUTS_FIRST_READ = true;// Track the first read

// Analog threshold states
var ANALOG_INPUTS_ILLEGAL = -1;
var ANALOG_INPUTS_IN_RANGE = 0;
var ANALOG_INPUTS_BELOW_THRESHOLD = 1;
var ANALOG_INPUTS_ABOVE_THRESHOLD = 2;
var AnalogThresholdValues = ["NORMAL", "BELOW", "ABOVE"];
//#endregion Analog Inputs

// Power Control
//#region Power Control
//  Power control offset will be determined by selected port #
var POWER_CONTROL_LOCATION = '/sys/class/gpio/gpio';
var POWER_CONTROL_BASEGPIO = 496;
//#endregion Power Control

//#endregion Constants (Device)

//#endregion Constants

// Device Class: Variables, Instantiation, Methods and Constructor
//#region Device Class: Variables, Instantiation, Methods and Constructor
function Device(port, logger, io, smtpManager, netSnmp) {

    // Class Variables
    //#region Class Variables

    // Class Variables (Common)
    //#region Class Variables (Common)
    var EventLog = [];
    var EnabledFlag = false;
    var SendEventsFlag = false;
    var DeviceModel;
    var ThisDeviceObject = this;
    //#endregion Class Variables (Common)

    // Class Variables (Device-specific)
    //#region Class Variables (Device-specific)
    // Digital input
    var DigitalCycleControl;// Manages the setInterval timer for digital I/O
    var DigitalInputsCycle;// Cycle speed
    var DigitalInterimUpdate;
    var DigitalSendInterimUpdate = false;
    var DigitalLastTimeBeforeInterimUpdate = moment();
    var DigitalCurrentTimeBeforeInterimUpdate;

    // Analog Input
    var AnalogCycleControl;// Manages the setInterval timer for analog I/O
    var AnalogInputsCycle;// Cycle time for analog I/O
    var AnalogInterimUpdate;
    var AnalogSendInterimUpdate = false;
    var AnalogLastTimeBeforeInterimUpdate = moment();
    var AnalogCurrentTimeBeforeInterimUpdate;
    //#endregion Class Variables (Device-specific)

    //#endregion Class Variables

    // Class Instantiation
    //#region Class Instantiation

    // Local Dependencies
    //#region Local Dependencies
    Logger = logger;
    SocketIo = io;
    SmtpManager = smtpManager;
    NetSnmp = netSnmp;
    //#endregion Local Dependencies

    try {
        // Load a private model for this instance of the device
        //  Remove model from cache to avoid overwrites
        var cache = require.resolve(DEVICE_MODEL);
        if (cache) delete require.cache[cache];
        //  Load model
        var DeviceModel = require(DEVICE_MODEL);
        //  Assign statics to model
        DeviceModel.port = port;
        DeviceModel.uuid = parseInt(moment().format("x"));
        Logger.trace(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".Constructor: DeviceModel=" + JSON.stringify(DeviceModel));
        Logger.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".Constructor: Device uuid=" + DeviceModel.uuid + " -----------------------");
    } catch (error) {
        Logger.error("Failed constructing device on port=" + port + " error=" + error);
    }
    //#endregion Class Instantiation

    // Class Methods
    //#region Class Methods

    //#region Class Methods (Common)

    //#region Class Methods (Common - Basic)
    this.setPort = function (port) {
        DeviceModel.port = port;
    }

    this.getPort = function () {
        return DeviceModel.port;
    }

    this.getDeviceId = function () {
        return DeviceModel.id;
    }

    this.getDeviceModel = function () {
        Logger.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".getDeviceModel: Device uuid=" + DeviceModel.uuid);
        return DeviceModel;
    }

    this.enable = function (enabled) {
        Logger.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".Enable: Enabled=" + (enabled ? "YES" : "NO"));
        if (enabled == EnabledFlag)
            return;// Don't do anything if we're already in the proper state
        else
            EnabledFlag = enabled;// Remember new setting

        if (enabled) {
            //#region Commission Device
            var power;
            if (DeviceModel.io.powerControl.enabled) power = 1; else power = 0;
            fs.truncate(POWER_CONTROL_LOCATION + (POWER_CONTROL_BASEGPIO + parseInt(DeviceModel.port)) + "/value", 0, function () {
                fs.writeFile(POWER_CONTROL_LOCATION + (POWER_CONTROL_BASEGPIO + parseInt(DeviceModel.port)) + "/value", power, function (err) {
                    if (err) {
                        return Logger.error(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".Enable: Power Set Error (Write)=" + JSON.stringify(err));
                    }
                    Logger.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".Enable: Port=" + DeviceModel.port + " - '" + POWER_CONTROL_LOCATION + (POWER_CONTROL_BASEGPIO + parseInt(DeviceModel.port)) + "/value" + "' Power=" + power);
                });
            });
            //#endregion Commission Device

            //#region Calculate Low and High Thresholds
            this.calculateThresholds();
            //#endregion Calculate Low and High Thresholds

            //#region Start Device Processing
            this.startProcessing();
            //#endregion Start Device Processing
        }

        else {
            //#region Stop Device Processing
            this.stopProcessing();
            //#endregion Stop Device Processing

            //#region Decommission Device
            var power;
            if (DeviceModel.io.powerControl.enabled) power = 0; else power = 1;
            fs.truncate(POWER_CONTROL_LOCATION + (POWER_CONTROL_BASEGPIO + parseInt(DeviceModel.port)) + "/value", 0, function () {
                fs.writeFile(POWER_CONTROL_LOCATION + (POWER_CONTROL_BASEGPIO + parseInt(DeviceModel.port)) + "/value", power, function (err) {
                    if (err) {
                        return Logger.error(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".Constructor: Power Set Error (Write)=" + JSON.stringify(err));
                    }
                    Logger.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".Constructor: Port=" + DeviceModel.port + " - '" + POWER_CONTROL_LOCATION + (POWER_CONTROL_BASEGPIO + parseInt(DeviceModel.port)) + "/value" + "' Power=" + power);
                });
            });
            //#endregion Decommission Device
        }
    }

    this.sendEvents = function (sendevents) {
        SendEventsFlag = sendevents;
    }
    //#endregion Class Methods (Common - Basic)

    //#region Class Methods (Common - UpdateDevice)
    this.updateDevice = function (port, model, callback) {
        Logger.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".updateDevice: Device uuid=" + DeviceModel.uuid);
        DeviceModel = model;
        this.calculateThresholds();// Update thresholds - specific to this device!
        callback(DeviceModel);
    }
    //#endregion Class Methods (Common - UpdateDevice)

    //#endregion Class Methods (Common)

    //#region Class Methods (Device Processing)

    //#region Class Methods (Device Processing - Start/Stop)
    //#region startProcessing
    this.startProcessing = function () {

        // Analog Inputs Cycle
        //AnalogInputsCycle = 5000;// Temporary
        AnalogInputsCycle = DeviceModel.io.analogIo.cycleTime;
        AnalogInterimUpdate = AnalogInputsCycle * 5;
        AnalogCycleControl = setInterval(function () { ThisDeviceObject.processAnalogInput() }, AnalogInputsCycle);

        // Digital Inputs Cycle
        //DigitalInputsCycle = 5000;// Temporary
        DigitalInputsCycle = DeviceModel.io.digitalIo.cycleTime;
        DigitalInterimUpdate = DigitalInputsCycle * 5;
        DeviceModel.parameters.PTT.HoldTimer = 0;
        DeviceModel.parameters.PTT.Value = -1;
        DigitalCycleControl = setInterval(function () { ThisDeviceObject.processDigitalInput() }, DigitalInputsCycle);

        Logger.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".startProcessing: IO processing cycle started");
    }
    //#endregion startProcessing

    //#region stopProcessing
    this.stopProcessing = function () {

        //  Stop analog and digital processing
        try {
            if (AnalogCycleControl) clearInterval(AnalogCycleControl);
            if (DigitalCycleControl) clearInterval(DigitalCycleControl);
            Logger.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".stopProcessing: Stopped IO processing cycle.");
        } catch (error) {
            Logger.error(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".stopProcessing: Stopped IO processing cycle. Error=" + error);
        }
    }
    //#endregion stopProcessing

    //#endregion Class Methods (Device Processing - Start/Stop)

    //#region Class Methods (Device Processing - Device Information)
    //#region getDeviceInformation
    this.getDeviceInformation = function () {
        return DeviceModel.parameters.DeviceInformation;
    }
    //#endregion getDeviceInformation
    //#endregion Class Methods (Device Processing - Device Information)

    //#region Class Methods (Device Processing - Process I/O)

    //#region Analog I/O
    //#region processAnalogInput
    //  Reads the value of the analog input (Power), based on qualifier and hold time
    //  Updates UI based on I/O changes
    //  Updates UI and SNMP/SMTP based on threshold changes
    this.processAnalogInput = function () {

        var timeChanged = moment();
        // This is to allow interim updates in case the user has changed screens between udpates
        AnalogCurrentTimeBeforeInterimUpdate = timeChanged;
        if ((AnalogCurrentTimeBeforeInterimUpdate - AnalogLastTimeBeforeInterimUpdate) > AnalogInterimUpdate) {
            AnalogSendInterimUpdate = true;
            AnalogLastTimeBeforeInterimUpdate = AnalogCurrentTimeBeforeInterimUpdate;
        } else {
            AnalogSendInterimUpdate = false;
        }

        try {
            // Read logic
            //  - Reads the data from the appropriate location
            //  - Removes any excess characters
            //  - Adjusts for offsets (optional)
            //  - Allocates to temporary value, pending filtering, threshold, qualification and publication logic
            //#region Read logic
            var rawValue = 0;
            var data = 0;
            var offset = parseInt(parseInt(DeviceModel.port) * 2 + parseInt(DeviceModel.io.analogIo.analog0.offset));
            try {
                Logger.trace(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".processAnalogInput: [Read: ACCESSIO] Reading from '" + ANALOG_INPUTS_LOCATION + "/in" + offset + "_input'");
                //Logger.warn(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".processAnalogInput: [Read: ACCESSIO] Reading from '" + ANALOG_INPUTS_LOCATION + "/in" + offset + "_input'");
                data = fs.readFileSync(ANALOG_INPUTS_LOCATION + "/in" + offset + "_input", 'utf8');
                //Logger.warn(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".processAnalogInput: [Read: ACCESSIO] Read data='" + data + "'");
            } catch (err) {
                // This is to catch infrequent errors related to reading
                Logger.error(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".processAnalogInput: [Read: ACCESSIO] Error=" + err);
                return;
            }

            // Get rid of any offending extra characters in read
            rawValue = String(data).replace(/(\r\n|\n|\r|[ +])/img, "");
            //rawValue = parseInt(rawValue) + Math.floor(Math.random() * 1001);//*** RANDOM
            // Add the offset to the read value
            rawValue = Math.round(Number(rawValue) + Number(this.convertValueToRawValue(DeviceModel.parameters.Power.Offset)) - Number(this.convertValueToRawValue(0)));
            // Create temp value to track allow us to handle hold times & qualifiers
            var tempAnalogInputPoint = JSON.parse(JSON.stringify(DeviceModel.parameters.Power));
            tempAnalogInputPoint.RawValue = rawValue;
            tempAnalogInputPoint.RangedValue = this.convertRawValueToValue(rawValue);
            //Logger.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".processAnalogInput: [Read: PARSEDIO] Current=" + rawValue + " CurrentRanged=" + parseFloat(tempAnalogInputPoint.RangedValue).toFixed(2) + "V");
            //Logger.warn(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".processAnalogInput: [Read: PARSEDIO] Current=" + rawValue + " CurrentRanged=" + parseFloat(tempAnalogInputPoint.RangedValue).toFixed(2) + "V");
            //#endregion Read logic

            // Filter logic
            //  - We ignore minute changes in the analog values, 
            //    change in raw value must be greater than ANALOG_INPUTS_CHANGE_FILTER
            //  - But we need to go through the logic when a HoldTimer is being used
            //#region Filter logic
            // Don't do anything if:
            //  - there has been no change in the raw value reading (so there's no real change in the input)
            //  - there is no hold time (so we don't have to start any timers)
            //  - the threshold state is valid (it hasn't been invalidated because of a configuration change)
            var valueChanged = Math.abs(DeviceModel.parameters.Power.RawValue - rawValue);
            if ((valueChanged <= ANALOG_INPUTS_CHANGE_FILTER)
                && (DeviceModel.parameters.Power.HoldTime == 0)
                && (DeviceModel.parameters.Power.ThresholdState >= 0)
                && !AnalogSendInterimUpdate) {
                return;
            }

            if (DeviceModel.parameters.Power.HoldTime > 0) {
                Logger.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".processAnalogInput: [Filter: HOLDTIMER=" + parseFloat(DeviceModel.parameters.Power.HoldTime).toFixed(2) + "s] Previous=" + DeviceModel.parameters.Power.RawValue + " Current=" + rawValue);
            } else {
                Logger.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".processAnalogInput: [Filter: VALUECHANGE] Previous=" + DeviceModel.parameters.Power.RawValue + " Current=" + rawValue);
            }
            //#endregion Filter logic

            // Threshold Change logic
            //  - ThresholdChanged = true or false, depending if we've changed a threshold
            //#region Threshold logic
            var ThresholdChanged = false;
            var oldThreshold = DeviceModel.parameters.Power.ThresholdState;
            var newThreshold = this.getAnalogInputThresholdState(tempAnalogInputPoint, tempAnalogInputPoint.RawValue);
            if ((oldThreshold != newThreshold) || (oldThreshold < 0)) {
                ThresholdChanged = true;
                Logger.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".processAnalogInput: [Threshold: CHANGED] PreviousRanged=" + parseFloat(DeviceModel.parameters.Power.RangedValue).toFixed(2) + "V (" + (oldThreshold != -1 ? AnalogThresholdValues[oldThreshold] : "UNDEFINED") + "),  CurrentRanged=" + tempAnalogInputPoint.RangedValue.toFixed(2) + "V (" + (newThreshold != -1 ? AnalogThresholdValues[newThreshold] : "UNDEFINED") + ")");
                //Logger.warn(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".processAnalogInput: [Threshold: CHANGED] PreviousRanged=" + parseFloat(DeviceModel.parameters.Power.RangedValue).toFixed(2) + "V (" + (oldThreshold != -1 ? AnalogThresholdValues[oldThreshold] : "UNDEFINED") + "),  CurrentRanged=" + tempAnalogInputPoint.RangedValue.toFixed(2) + "V (" + (newThreshold != -1 ? AnalogThresholdValues[newThreshold] : "UNDEFINED") + ")");
            } else {
                Logger.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".processAnalogInput: [Threshold: UNCHANGED] PreviousRanged=" + parseFloat(DeviceModel.parameters.Power.RangedValue).toFixed(2) + "V (" + (oldThreshold != -1 ? AnalogThresholdValues[oldThreshold] : "UNDEFINED") + "),  CurrentRanged=" + tempAnalogInputPoint.RangedValue.toFixed(2) + "V (" + (newThreshold != -1 ? AnalogThresholdValues[newThreshold] : "UNDEFINED") + ")");
                //Logger.warn(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".processAnalogInput: [Threshold: UNCHANGED] PreviousRanged=" + parseFloat(DeviceModel.parameters.Power.RangedValue).toFixed(2) + "V (" + (oldThreshold != -1 ? AnalogThresholdValues[oldThreshold] : "UNDEFINED") + "),  CurrentRanged=" + tempAnalogInputPoint.RangedValue.toFixed(2) + "V (" + (newThreshold != -1 ? AnalogThresholdValues[newThreshold] : "UNDEFINED") + ")");
            }
            //#endregion Threshold logic

            // Qualification logic
            //  - QualifierValid = true or false, depending if qualified by digital input
            //  - DeviceModel.parameters.Power..HoldTimerActive = true or false, depending on qualifications status
            //  - DeviceModel.parameters.Power..HoldTimer = holds time of when first qualified, with or without qualifier
            //#region Qualification logic
            var QualifierValid = false;
            if (DeviceModel.parameters.Power.Qualifier) {
                // #region QUALIFIER USED
                if (DeviceModel.parameters.PTT.Value) {
                    QualifierValid = true;// Qualifier is valid ...

                    //#region HoldTime logic for Qualifier USED & TRUE
                    if (DeviceModel.parameters.Power.HoldTime > 0) {
                        // If there is a hold time ...
                        if (DeviceModel.parameters.Power.HoldTimerActive === false) {
                            // If the hold timer was previously not running ...
                            if (ThresholdChanged) {
                                //  Initial Threshold Change
                                //#region Initial Threshold Change
                                DeviceModel.parameters.Power.HoldTimer = timeChanged;
                                DeviceModel.parameters.Power.HoldTimerActive = true;
                                DeviceModel.parameters.Power.HoldTimeState = HOLDSTATE_HOLDING;
                                this.sendSocket(DeviceModel.port, DeviceModel.uuid, "HoldTimeState", DeviceModel.parameters.Power);
                                Logger.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".processAnalogInput: [Qualifier(YES) + HoldTime: STARTED] HoldTimer=0.00");
                                //#endregion Initial Threshold Change
                            }
                        }
                        else {
                            // If the hold timer was previously running ...
                            if (!ThresholdChanged) {
                                // Threshold Changed, but HoldTimer was previously holding
                                //#region Threshold Changed, but HoldTimer was previously holding
                                if (DeviceModel.parameters.Power.HoldTimerActive == true) {
                                    //  And the threshold has returned to same place ... reset the timer
                                    DeviceModel.parameters.Power.HoldTimerActive = false;
                                    DeviceModel.parameters.Power.HoldTimeState = HOLDSTATE_OFF;
                                    this.sendSocket(DeviceModel.port, DeviceModel.uuid, "HoldTimeState", DeviceModel.parameters.Power);
                                    Logger.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".processAnalogInput: [Qualifier(YES) + HoldTime: RESET] HoldTimer=0.00");
                                }
                                //#endregion Threshold Changed, but HoldTimer was previously holding
                            }
                        }
                    }
                    //#endregion HoldTime logic for Qualifier USED & TRUE
                } else {
                    QualifierValid = false;// Qualifier is invalid ...

                    //#region HoldTime logic for Qualifier USED & FALSE
                    if (DeviceModel.parameters.Power.HoldTime > 0) {
                        // Reset HoldTime due to failed Qualifier
                        //#region Reset HoldTime due to failed Qualifier
                        if (DeviceModel.parameters.Power.HoldTimerActive == true) {
                            // Reset hold timer ... it will have to be requalified
                            DeviceModel.parameters.Power.HoldTimerActive = false;
                            DeviceModel.parameters.Power.HoldTimeState = HOLDSTATE_OFF;
                            this.sendSocket(DeviceModel.port, DeviceModel.uuid, "HoldTimeState", DeviceModel.parameters.Power);
                            Logger.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".processAnalogInput: [Qualifier(FALSE) + HoldTime: RESET] HoldTimer=0.00");
                            Logger.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".processAnalogInput: [HOLD TIMER - HoldTimer=" + (DeviceModel.parameters.Power.HoldTime > 0 ? "RESET (Qualifier Failed)" : "NOT_USED") + ", Qualifier=USED, QualifierValid=FALSE] '" + DeviceModel.parameters.Power.Id + "' using qualifier=DigitalInput[" + (DeviceModel.parameters.Power.Qualifier) + "]");
                        }
                        //#endregion Reset HoldTime due to failed Qualifier
                    }
                    //#endregion HoldTime logic for Qualifier USED & FALSE
                }
                // #endregion QUALIFIER USED
            }
            else {
                //#region QUALIFIER NOT USED
                QualifierValid = true;
                if (DeviceModel.parameters.Power.HoldTime > 0) {
                    // If there is a hold time ...
                    if (DeviceModel.parameters.Power.HoldTimerActive === false) {
                        // If the hold timer was previously not running ...
                        if (ThresholdChanged) {
                            // Initial Threshold Change
                            //#region Initial Threshold Change
                            DeviceModel.parameters.Power.HoldTimer = timeChanged;
                            DeviceModel.parameters.Power.HoldTimerActive = true;
                            DeviceModel.parameters.Power.HoldTimeState = HOLDSTATE_HOLDING;
                            this.sendSocket(DeviceModel.port, DeviceModel.uuid, "HoldTimeState", DeviceModel.parameters.Power);
                            Logger.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".processAnalogInput: [Qualifier(NOTUSED) + HoldTime: STARTED] HoldTimer=0.00");
                            //#endregion Initial Threshold Change
                        }
                    }
                    else {
                        //  And the threshold has returned to same place ... reset the timer
                        // If the hold timer was previously running ...
                        // Reset HoldTime
                        //#region Reset HoldTime
                        if (!ThresholdChanged) {
                            if (DeviceModel.parameters.Power.HoldTimerActive == true) {
                                DeviceModel.parameters.Power.HoldTimerActive = false;
                                DeviceModel.parameters.Power.HoldTimeState = HOLDSTATE_OFF;
                                this.sendSocket(DeviceModel.port, DeviceModel.uuid, "HoldTimeState", DeviceModel.parameters.Power);
                                Logger.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".processAnalogInput: [Qualifier(NOTUSED) + HoldTime: RESET] HoldTimer=0.00");
                            }
                        }
                        //#endregion Reset HoldTime
                    }
                }
                //#endregion HoldTime logic for QUALIFIER NOT USED
            }
            //#endregion Qualification logic

            // HoldTime logic
            //  - HoldTimeValid = true or false, depending on:
            //      - true: if holdtimer is used & >= holdtime
            //      - true: if holdtimer is not used
            //      - false: all other conditions
            //#region HoldTime logic
            var HoldTimeValid = false;
            var HoldTimeChanged = 0;
            var HoldTimeActive = false;
            if (DeviceModel.parameters.Power.HoldTime > 0) {
                // Using a hold time ...
                if (DeviceModel.parameters.Power.HoldTimerActive == true) {
                    // Hold time is valid ...
                    HoldTimeActive = true;
                    HoldTimeChanged = (timeChanged - DeviceModel.parameters.Power.HoldTimer) / 1000;
                    if (HoldTimeChanged >= DeviceModel.parameters.Power.HoldTime) {
                        HoldTimeValid = true;
                        Logger.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".processAnalogInput: [HoldTime: ACTIVE] HoldTimer=" + parseFloat(HoldTimeChanged).toFixed(2));
                        //Logger.warn(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".processAnalogInput: [HoldTime: ACTIVE] HoldTimer=" + parseFloat(HoldTimeChanged).toFixed(2));
                        // Turn on, then turn off Hold Time button
                        DeviceModel.parameters.Power.HoldTimeState = HOLDSTATE_ACTIVE;
                        this.sendSocket(DeviceModel.port, DeviceModel.uuid, "HoldTimeState", DeviceModel.parameters.Power);
                        DeviceModel.parameters.Power.HoldTimeState = HOLDSTATE_OFF;
                        this.sendSocket(DeviceModel.port, DeviceModel.uuid, "HoldTimeState", DeviceModel.parameters.Power);
                    } else {
                        DeviceModel.parameters.Power.HoldTimeState = HOLDSTATE_HOLDING;
                        this.sendSocket(DeviceModel.port, DeviceModel.uuid, "HoldTimeState", DeviceModel.parameters.Power);
                        Logger.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".processAnalogInput: [HoldTime: HOLDING] HoldTimer=" + parseFloat(HoldTimeChanged).toFixed(2));
                        //Logger.warn(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".processAnalogInput: [HoldTime: HOLDING] HoldTimer=" + parseFloat(HoldTimeChanged).toFixed(2));
                    }
                }
            } else {
                // Not using a hold time ...
                HoleTimeActive = false;
                HoldTimeValid = true;
                Logger.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".processAnalogInput: [HoldTime: NOTUSED]");
            }
            //#endregion HoldTime logic

            // Publish logic (real-time info and SNMP traps)
            //  Refer to the 'Crest - Real-time updates and SNMP traps.xlsx' document for cases
            //#region Publish logic
            var FormulaIndex = DeviceModel.device.FormulaIndex;
            if (QualifierValid && HoldTimeValid) {
                // If the change is qualified and holdtime is valid ...
                //#region QualifierValid=TRUE and HoldTimeValid=TRUE preamble
                // Update the local array to the new value

                // Calculate Formula info, if present
                DeviceModel.parameters.Power.RawValue = rawValue;
                DeviceModel.parameters.Power.RangedValue = this.convertRawValueToValue(DeviceModel.parameters.Power.RawValue).toFixed(2);// Update the Ranged Value
                this.calculateFormula(DeviceModel.parameters.Power.RangedValue);// Calculates all formulas
                //#endregion QualifierValid=TRUE and HoldTimeValid=TRUE preamble

                if (ThresholdChanged) {
                    // Threshold change is qualified and hold time is valid, so record the changes
                    //  *** CASE 1 : QualifierValid=TRUE, HoldTimeValid=TRUE, ThresholdChange=TRUE
                    //  *** CASE 2B: QualifierValid=TRUE, HoldTimeValid=TRUE, ThresholdChange=TRUE
                    //  *** CASE 3A: QualifierValid=TRUE, HoldTimeValid=TRUE, ThresholdChange=TRUE
                    //  *** CASE 4D: QualifierValid=TRUE, HoldTimeValid=TRUE, ThresholdChange=TRUE
                    //#region QualifierValid=TRUE, HoldTimeValid=TRUE, ThresholdChange=TRUE
                    DeviceModel.parameters.Power.TimeChanged = timeChanged;
                    DeviceModel.parameters.Power.ThresholdState = newThreshold;
                    DeviceModel.parameters.Power.HoldTimerActive = false;// If the threshold change, set HoldTimerState back to false to restart HoldTimer

                    // Update real-time value, including SNMP because of threshold change
                    //  If this is the first time through, report only if NORMAL state
                    //  All other times report do a report regardless of state
                    Logger.trace(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".processAnalogInput: [Publish: UPDATETHRESHOLD] ANALOG_INPUTS_FIRST_READ=" + ANALOG_INPUTS_FIRST_READ + " ThresholdState=" + DeviceModel.parameters.Power.ThresholdState);
                    if ((ANALOG_INPUTS_FIRST_READ == false) || (ANALOG_INPUTS_FIRST_READ == true && DeviceModel.parameters.Power.ThresholdState != ANALOG_INPUTS_IN_RANGE)) {
                        this.sendEvent(DeviceModel.port, DeviceModel.uuid, DeviceModel.parameters.Power);
                    }

                    // ***** Record the analog input change in the database

                    Logger.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".processAnalogInput: [Publish: UPDATETHRESHOLD] Current=" + DeviceModel.parameters.Power.RawValue + " CurrentRanged=" + DeviceModel.parameters.Power.RangedValue + " V" + (FormulaIndex !== -1 ? " FormulaValue=" + DeviceModel.parameters.Power.Formulas[DeviceModel.parameters.Power.FormulaIndex].FormulaValue : "") + " Threshold=" + (newThreshold != -1 ? AnalogThresholdValues[newThreshold] : "UNDEFINED") + " HoldTime=" + (DeviceModel.parameters.Power.HoldTime == 0 ? "0" : HoldTimeChanged.toFixed(2)) + "s");
                    Logger.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".processAnalogInput: [Publish: UPDATETHRESHOLD] Power=" + JSON.stringify(DeviceModel.parameters.Power));
                    //#endregion QualifierValid=TRUE, HoldTimeValid=TRUE, ThresholdChange=TRUE
                }
                else {
                    // No threshold changed, so just send real time update
                    //  *** CASE 1 : QualifierValid=TRUE, HoldTimeValid=TRUE, ThresholdChange=FALSE
                    //  *** CASE 2B: QualifierValid=TRUE, HoldTimeValid=TRUE, ThresholdChange=FALSE
                    //  *** CASE 3A: QualifierValid=TRUE, HoldTimeValid=TRUE, ThresholdChange=FALSE
                    //  *** CASE 4D: QualifierValid=TRUE, HoldTimeValid=TRUE, ThresholdChange=FALSE
                    //#region QualifierValid=TRUE, HoldTimeValid=TRUE, ThresholdChange=FALSE

                    // Update real-time value, if there was value change or we're holding
                    if (valueChanged > ANALOG_INPUTS_CHANGE_FILTER) {
                        this.sendSocket(DeviceModel.port, DeviceModel.uuid, "Value", DeviceModel.parameters.Power);
                        Logger.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".processAnalogInput: [Publish: UPDATEVALUE] Current=" + DeviceModel.parameters.Power.RawValue + " CurrentRanged=" + DeviceModel.parameters.Power.RangedValue + " V" + (FormulaIndex !== -1 ? " FormulaValue=" + DeviceModel.parameters.Power.Formulas[DeviceModel.parameters.Power.FormulaIndex].FormulaValue : "") + " Threshold=" + (newThreshold != -1 ? AnalogThresholdValues[newThreshold] : "UNDEFINED") + " HoldTime=" + (DeviceModel.parameters.Power.HoldTime == 0 ? "0" : HoldTimeChanged.toFixed(2)) + "s");
                    }
                    else if (AnalogSendInterimUpdate) {
                        this.sendSocket(DeviceModel.port, DeviceModel.uuid, "Value", DeviceModel.parameters.Power);
                        Logger.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".processAnalogInput: [Publish: UPDATEINTERIM] Current=" + DeviceModel.parameters.Power.RawValue + " CurrentRanged=" + DeviceModel.parameters.Power.RangedValue + " V" + (FormulaIndex !== -1 ? " FormulaValue=" + DeviceModel.parameters.Power.Formulas[DeviceModel.parameters.Power.FormulaIndex].FormulaValue : "") + " Threshold=" + (newThreshold != -1 ? AnalogThresholdValues[newThreshold] : "UNDEFINED") + " HoldTime=" + (DeviceModel.parameters.Power.HoldTime == 0 ? "0" : HoldTimeChanged.toFixed(2)) + "s");
                    }
                    //#endregion QualifierValid=TRUE, HoldTimeValid=TRUE, ThresholdChange=FALSE
                }
            } else {

                // If we're qualified and changes are happening within the threshold value, send an update
                if (QualifierValid) {
                    // No threshold changed, so just send real time update
                    //  *** CASE 2A: QualifierValid=TRUE, HoldTimeValid=FALSE, ValueChange=TRUE, ThresholdChange=FALSE
                    //  *** CASE 2A: QualifierValid=TRUE, HoldTimeValid=FALSE, ValueChange=TRUE, ThresholdChange=TRUE
                    //  *** CASE 4C: QualifierValid=TRUE, HoldTimeValid=FALSE, ValueChange=TRUE, ThresholdChange=FALSE
                    //  *** CASE 4C: QualifierValid=TRUE, HoldTimeValid=FALSE, ValueChange=TRUE, ThresholdChange=TRUE
                    //#region QualifierValid=TRUE, HoldTimeValid=FALSE, ValueChange=TRUE, ThresholdChange=DONTCARE
                    DeviceModel.parameters.Power.RawValue = rawValue;
                    DeviceModel.parameters.Power.RangedValue = this.convertRawValueToValue(DeviceModel.parameters.Power.RawValue).toFixed(2);// Update the Ranged Value
                    this.calculateFormula(DeviceModel.parameters.Power.RangedValue);// Calculates all formulas

                    // Update real-time value, if there was value change or we're holding
                    if (valueChanged > ANALOG_INPUTS_CHANGE_FILTER) {
                        this.sendSocket(DeviceModel.port, DeviceModel.uuid, "Value", DeviceModel.parameters.Power);
                        Logger.trace(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".processAnalogInput: [REAL-TIME UPDATE - QualifierValid=" + QualifierValid + ", HoldTimeValid=" + HoldTimeValid + "] '" + DeviceModel.parameters.Power.Id + "' ValueChanged=" + valueChanged + " RawValue=" + DeviceModel.parameters.Power.RawValue + " RangedValue=" + DeviceModel.parameters.Power.RangedValue + " V" + (FormulaIndex !== -1 ? " FormulaValue=" + DeviceModel.parameters.Power.Formulas[DeviceModel.parameters.Power.FormulaIndex].FormulaValue : "") + " Threshold=" + (newThreshold != -1 ? AnalogThresholdValues[newThreshold] : "UNDEFINED"));
                    }
                    else if (AnalogSendInterimUpdate) {
                        // We won't clog up the log file with this during debug
                        this.sendSocket(DeviceModel.port, DeviceModel.uuid, "Value", DeviceModel.parameters.Power);
                        Logger.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".processAnalogInput: [REAL-TIME UPDATE(Interim) - QualifierValid=" + QualifierValid + ", HoldTimeValid=" + HoldTimeValid + "] '" + DeviceModel.parameters.Power.Id + "' ValueChanged=" + valueChanged + " RawValue=" + DeviceModel.parameters.Power.RawValue + " RangedValue=" + DeviceModel.parameters.Power.RangedValue + " V" + (FormulaIndex !== -1 ? " FormulaValue=" + DeviceModel.parameters.Power.Formulas[DeviceModel.parameters.Power.FormulaIndex].FormulaValue : "") + " Threshold=" + (newThreshold != -1 ? AnalogThresholdValues[newThreshold] : "UNDEFINED"));
                    }
                    //#endregion QualifierValid=TRUE, HoldTimeValid=FALSE, ValueChange=TRUE, ThresholdChange=DONTCARE
                }
                else {
                    // Not qualified, don't do anyting ...
                    Logger.trace(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".processAnalogInput: [Publish: Qualifier(FALSE)]");
                }
            }
            //#endregion Publish logic
        }
        catch (error) {
            Logger.error(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".processAnalogInput: Process Error (General)=" + error);
            Logger.error(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".processAnalogInput: previousValue=" + DeviceModel.parameters.Power.RawValue + " rawValue=" + rawValue);
        }
        finally {
            ANALOG_INPUTS_FIRST_READ = false;// We've done our initial read
        }
    }
    //#endregion processAnalogInput

    // Convert a raw value to a ranged value
    //#region convertRawValueToValue
    this.convertRawValueToValue = function (rawValue) {
        var rangedValue = Number(parseFloat(rawValue) * ANALOG_INPUTS_RAWVALUE_SLOPE + ANALOG_INPUTS_RAWVALUE_OFFSET);
        if (rangedValue < 0) rangedValue = 0;
        return rangedValue;
    }
    //#endregion convertRawValueToValue

    // Convert a ranged value to raw value
    //#region convertValueToRawValue
    this.convertValueToRawValue = function (rangedValue) {
        var rawValue = Math.round((parseFloat(rangedValue) - ANALOG_INPUTS_RAWVALUE_OFFSET) / ANALOG_INPUTS_RAWVALUE_SLOPE);
        return rawValue;
    }
    //#endregion convertValueToRawValue

    // Get the threshold state
    //#region getAnalogInputThresholdState
    this.getAnalogInputThresholdState = function (AnalogInput, Value) {
        var LowThresholdRawValue = this.convertValueToRawValue(parseFloat(AnalogInput.RangedLowThreshold).toFixed(2));
        var HighThresholdRawValue = this.convertValueToRawValue(parseFloat(AnalogInput.RangedHighThreshold).toFixed(2));
        var CurrentRangedValue = this.convertRawValueToValue(Value);

        var ThresholdState = ANALOG_INPUTS_IN_RANGE;
        if (Value < LowThresholdRawValue) {
            // This specifically tests for the 0.00V case
            if (parseFloat(AnalogInput.RangedLowThreshold).toFixed(2) != parseFloat(CurrentRangedValue).toFixed(2)) {
                ThresholdState = ANALOG_INPUTS_BELOW_THRESHOLD
            } else {
                //Logger.info("Analog Inputs - [NO THRESHOLD CHANGE] Analog Input[" + AnalogInput.Address + "] LowThreshold=" + AnalogInput.RangedLowThreshold + " Value=" + CurrentRangedValue);
            }
        } else if (Value > HighThresholdRawValue) {
            ThresholdState = ANALOG_INPUTS_ABOVE_THRESHOLD
        }

        return ThresholdState;
    }
    //#endregion getAnalogInputThresholdState

    // Evaluate an existing Formula
    //#region Calculate Formulas and Thresholds
    this.calculateFormula = function (rangedValue) {

        // Evaluate Formulas
        var formulaResult;
        var i;
        try {
            for (i = 0; i < DeviceModel.parameters.Power.Formulas.length; i++) {
                var expression = mathparser.parse(DeviceModel.parameters.Power.Formulas[i].Formula);

                formulaResult = expression.evaluate({ x: rangedValue });
                DeviceModel.parameters.Power.Formulas[i].FormulaValue = Number(formulaResult).toFixed(DeviceModel.parameters.Power.Formulas[i].Decimals) + " " + DeviceModel.parameters.Power.Formulas[i].Units;
                //Logger.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" +".calculateFormula: Successful evaluation - result= '" + " " + DeviceModel.parameters.Power.Formulas[i].FormulaValue + "'");
            }
        }
        catch (err) {
            Logger.error(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".calculateFormula: Unsuccessful formula evaluation of '" + DeviceModel.parameters.Power.Formulas[i].Formula + "' err=" + err);
        }
    };

    this.calculateThresholds = function () {
        // Evaluate Formulas
        var lowResult;
        var highResult;
        var i;
        try {
            for (i = 0; i < DeviceModel.parameters.Power.Formulas.length; i++) {
                var expression = mathparser.parse(DeviceModel.parameters.Power.Formulas[i].Formula);

                lowResult = expression.evaluate({ x: parseFloat(DeviceModel.parameters.Power.RangedLowThreshold).toFixed(2) });
                DeviceModel.parameters.Power.Formulas[i].FormulaLowThreshold = Number(lowResult).toFixed(DeviceModel.parameters.Power.Formulas[i].Decimals) + " " + DeviceModel.parameters.Power.Formulas[i].Units;
                highResult = expression.evaluate({ x: parseFloat(DeviceModel.parameters.Power.RangedHighThreshold).toFixed(2) });
                DeviceModel.parameters.Power.Formulas[i].FormulaHighThreshold = Number(highResult).toFixed(DeviceModel.parameters.Power.Formulas[i].Decimals) + " " + DeviceModel.parameters.Power.Formulas[i].Units;
                Logger.trace(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".calculateFormula: Successful evaluation - low= '" + DeviceModel.parameters.Power.Formulas[i].FormulaLowThreshold + "'  high= '" + DeviceModel.parameters.Power.Formulas[i].FormulaHighThreshold + "'");
            }
        }
        catch (err) {
            Logger.error(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".calculateFormula: Unsuccessful evaluation of Low and High formula '" + DeviceModel.parameters.Power.Formulas[i].Formula + "' err=" + err);
        }
    }
    //#endregion Calculate Formulas and Thresholds    
    //#endregion Analog I/O 

    //#region Digital I/O 
    //#region processDigitalInput
    //  Reads the value of the digital input (PTT), based on hold time
    //  Updates UI and SNMP/SMTP based on state changes
    this.processDigitalInput = function () {

        var timeChanged = moment();
        var rawValue;

        // This is to allow interim updates in case the user has changed screens between udpates
        DigitalCurrentTimeBeforeInterimUpdate = timeChanged;
        if ((DigitalCurrentTimeBeforeInterimUpdate - DigitalLastTimeBeforeInterimUpdate) > DigitalInterimUpdate) {
            DigitalSendInterimUpdate = true;
            DigitalLastTimeBeforeInterimUpdate = DigitalCurrentTimeBeforeInterimUpdate;
        } else {
            DigitalSendInterimUpdate = false;
        }

        try {
            // Only read value if the qualifier is being used
            //Logger.warn(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" +".processDigitalInput: Entry for Port=" + DeviceModel.port + " Qualifier Used=" + DeviceModel.parameters.Power.Qualifier);
            if (DeviceModel.parameters.Power.Qualifier) {
                // Read Logic
                //  If the raw value has changed then we'll ahve to do stuff
                //#region Read Logic

                var rawValue = 0;
                var rangedValue = 0;
                var qualifierValue = false;
                var data = 0;
                var offset = parseInt(parseInt(DeviceModel.port) * 2 + parseInt(DeviceModel.io.analogIo.analog1.offset));
                try {
                    Logger.trace(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".processDigitalInput: [Read: ACCESSIO] Reading from '" + ANALOG_INPUTS_LOCATION + "/in" + offset + "_input'");
                    data = fs.readFileSync(ANALOG_INPUTS_LOCATION + "/in" + offset + "_input", 'utf8');
                } catch (err) {
                    // This is to catch infrequent errors related to reading
                    Logger.error(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".processDigitalInput: [Read: ACCESSIO] Error=" + err);
                    return;
                }

                // Get rid of any offending extra characters in read
                rawValue = String(data).replace(/(\r\n|\n|\r|[ +])/img, "");
                //rawValue = parseInt(rawValue) + Math.floor(Math.random() * 1001);//*** RANDOM
                // Add the offset to the read value
                rawValue = Math.round(Number(rawValue) + Number(this.convertValueToRawValue(DeviceModel.parameters.Power.Offset)) - Number(this.convertValueToRawValue(0)));
                rangedValue = this.convertRawValueToValue(rawValue);

                // Normally Closed vs Normally Open
                if (DeviceModel.parameters.PTT.Mode == "Normally Open") {
                    if (rangedValue > ANALOG_INPUTS_TTL_ONOFF) {
                        qualifierValue = true;
                    } else {
                        qualifierValue = false;
                    }
                } else {
                    if (rangedValue > ANALOG_INPUTS_TTL_ONOFF) {
                        qualifierValue = false;
                    } else {
                        qualifierValue = true;
                    }
                }
                Logger.trace(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".processDigitalInput: qualifierValue=" + qualifierValue);

                if ((DeviceModel.parameters.PTT.Value != qualifierValue) || (DeviceModel.parameters.PTT.HoldTime != 0)) {
                    // Filter logic - we only consider changes if there was a:
                    //  - Change in the digital value, or
                    //  - A hold timer is being used
                    //#region Filter logic                  
                    //if (DeviceModel.parameters.PTT.HoldTime > 0) {
                    //    Logger.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" +".processDigitalInput: [Filter: HOLDTIMER=" + parseFloat(DeviceModel.parameters.PTT.HoldTime).toFixed(2) + "s] Previous=" + DeviceModel.parameters.PTT.Value + " Current=" + rawValue);
                    //    Logger.warn(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" +".processDigitalInput: [Filter: HOLDTIMER=" + parseFloat(DeviceModel.parameters.PTT.HoldTime).toFixed(2) + "s] Previous=" + DeviceModel.parameters.PTT.Value + " Current=" + rawValue);
                    //} else {
                    //    Logger.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" +".processDigitalInput: [Filter: VALUECHANGE] Previous Value=" + DeviceModel.parameters.PTT.Value + " Current=" + qualifierValue);
                    //    Logger.warn(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" +".processDigitalInput: [Filter: VALUECHANGE] Previous Value=" + DeviceModel.parameters.PTT.Value + " Current=" + qualifierValue);
                    //}
                    //#endregion Filter logic

                    // HoldTime logic
                    //  - A non-zero value in the HoldTimer variable means that the HoldTimer is active
                    //  - Start the timer if there was a change and the timer wasn't previously started
                    //  - If the timer was previously started, then check if the 
                    //#region HoldTime logic
                    var HoldTimeValid = false;
                    var HoldTimeChanged = 0;
                    if (DeviceModel.parameters.PTT.HoldTime > 0) {
                        // HoldTime is used
                        //#region HoldTime is used
                        if (DeviceModel.parameters.PTT.Value != qualifierValue) {
                            // Value has changed ...
                            if (DeviceModel.parameters.PTT.HoldTimer === 0) {
                                // HoldTimer not started, so start it
                                //#region HoldTime off, start timer
                                DeviceModel.parameters.PTT.HoldTimer = timeChanged;// Start the timer
                                DeviceModel.parameters.PTT.HoldTimeState = HOLDSTATE_HOLDING;
                                this.sendSocket(DeviceModel.port, DeviceModel.uuid, "HoldTimeState", DeviceModel.parameters.PTT);
                                Logger.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".processDigitalInput: [HoldTime: STARTED] HoldTimer=0.00");
                                //#endregion HoldTime off, start timer
                            }
                            else {
                                HoldTimeChanged = (timeChanged - DeviceModel.parameters.PTT.HoldTimer) / 1000;
                                if (HoldTimeChanged >= DeviceModel.parameters.PTT.HoldTime) {
                                    // HoldTime active and reached
                                    //#region Holdtime active and reached
                                    HoldTimeValid = true;
                                    DeviceModel.parameters.PTT.HoldTimeState = (qualifierValue ? HOLDSTATE_ACTIVE : HOLDSTATE_OFF);
                                    this.sendSocket(DeviceModel.port, DeviceModel.uuid, "HoldTimeState", DeviceModel.parameters.PTT);
                                    Logger.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".processDigitalInput: [HoldTime: ACTIVE] HoldTimer=" + parseFloat(HoldTimeChanged).toFixed(2));
                                    DeviceModel.parameters.PTT.HoldTimer = 0;// Set hold timer to 0
                                    //#endregion Holdtime active and reached
                                } else {
                                    // HoldTime active, but not reached
                                    //#region HoldTime active, but not reached
                                    DeviceModel.parameters.PTT.HoldTimeState = HOLDSTATE_HOLDING;
                                    this.sendSocket(DeviceModel.port, DeviceModel.uuid, "HoldTimeState", DeviceModel.parameters.PTT);
                                    Logger.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".processDigitalInput: [HoldTime: HOLDING] HoldTimer=" + parseFloat(HoldTimeChanged).toFixed(2));
                                    //#endregion HoldTime active, but not reached
                                }
                            }
                        }
                        else {
                            // Value has either returned to the same value or hasn't changed, so reset the hold timer
                            //#region HoldTime is reset
                            if (DeviceModel.parameters.PTT.HoldTimer !== 0) {
                                // Returned to the same value
                                DeviceModel.parameters.PTT.HoldTimer = 0;
                                DeviceModel.parameters.PTT.HoldTimeState = (qualifierValue ? HOLDSTATE_ACTIVE : HOLDSTATE_OFF);
                                this.sendSocket(DeviceModel.port, DeviceModel.uuid, "HoldTimeState", DeviceModel.parameters.PTT);
                                Logger.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".processDigitalInput: [HoldTime: RESET] HoldTimer=" + parseFloat(HoldTimeChanged).toFixed(2));
                            }
                            //#endregion HoldTime is reset
                        }
                        //#endregion HoldTime is used
                    } else {
                        // HoldTime is not used
                        //#region HoldTime is not used
                        HoldTimeValid = true;
                        DeviceModel.parameters.PTT.HoldTimeState = (qualifierValue ? HOLDSTATE_ACTIVE : HOLDSTATE_OFF);
                        Logger.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".processDigitalInput: [HoldTime: NOT USED]");
                        //#endregion HoldTime is not used
                    }
                    //#endregion HoldTime logic

                    // Publish logic
                    //  - If the value has changed and the hold timer logic has been met, then change the value            
                    //#region Publish logic
                    if ((DeviceModel.parameters.PTT.Value != qualifierValue) && HoldTimeValid) {
                        // Change is valid ... 
                        //#region Value has changed
                        DeviceModel.parameters.PTT.Value = qualifierValue;
                        DeviceModel.parameters.PTT.TimeChanged = timeChanged;

                        // Update real-time value, including SNMP because of threshold change
                        //SendEvent(DeviceModel.port, DeviceModel.uuid, DeviceModel.parameters.PTT);// We don't need to report this for this device, not an alarm issue
                        this.sendSocket(DeviceModel.port, DeviceModel.uuid, "Value", DeviceModel.parameters.PTT);

                        Logger.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".processDigitalInput: [Publish: UPDATE] PTT=" + DeviceModel.parameters.PTT.Value + " (HoldTime=" + (DeviceModel.parameters.PTT.HoldTime == 0 ? "Not Used)" : parseFloat(HoldTimeChanged).toFixed(2) + "s)"));
                        Logger.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".processDigitalInput: [Publish: UPDATE] PTT=" + JSON.stringify(DeviceModel.parameters.PTT));

                        //#endregion Value has changed
                    }
                    //#endregion Publish logic
                }
            } // try { Processing }
        }// if (DeviceModel.parameters.PTT.Enabled === "YES")
        catch (error) {
            Logger.error(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".processDigitalInput: Process Error=" + error);
        }
        finally {
            // Nothing here for now ...
        }
    }
    //#endregion processDigitalInput
    //#endregion Digital I/O 

    //#endregion Process I/O
    //#endregion Class Methods (Device Processing - Process I/O)
    //#endregion Class Methods (Device Processing)

    //#region Class Methods (Eventing)

    //#region Class Methods (Eventing - SendEvent)
    this.sendEvent = function (port, uuid, parameter) {
        this.sendSocket(port, uuid, "Value", parameter)
        this.sendSMTP(port, uuid, parameter);
        this.sendSNMP(port, uuid, parameter);

        // EventLog
        //#region EventLog

        // Retrieve EventLog array
        if (fs.existsSync("log/port" + port + "events.json")) {
            fs.readFile("log/port" + port + "events.json", function (error, eventlog) {
                if (error) {
                    EventLog = [];// Error in EventLog retrieval
                    Logger.error(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".sendEvent: EventLog retrieve error. Error=" + error);
                } else {
                    EventLog = JSON.parse(eventlog);
                    Logger.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".sendEvent: EventLog retrieved. EventLog=" + JSON.stringify(EventLog));
                }

                // Update EventLog with latest event
                EventLog.unshift(parameter);// To the front of the array
                if (EventLog.length > EVENTS_LENGTH) {
                    EventLog.splice(-1);// Remove from the end if limit reached
                }

                // Save updated EventLog array          
                fs.truncate("log/port" + port + "events.json", 0, function () {
                    fs.writeFile("log/port" + port + "events.json", JSON.stringify(EventLog), function (error) {
                        if (error) {
                            Logger.error(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".sendEvent: EventLog save error. Error=" + error);
                        } else {
                            Logger.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".sendEvent: EventLog updated and saved. EventLog=" + JSON.stringify(EventLog));
                            cp.exec("sync", function (error, stdout, stderr) {
                                Logger.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".sendEvent: sync stdout=" + stdout);
                                if (error) {
                                    Logger.error(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".sendEvent: sync error=" + error);
                                }
                            });
                        }
                    });
                });

            });
        } else {
            // Update EventLog with latest event
            EventLog = [];// Empty event log
            EventLog.unshift(parameter);// To the front of the array

            // Save updated EventLog array
            fs.truncate("log/port" + port + "events.json", 0, function () {
                fs.writeFile("log/port" + port + "events.json", JSON.stringify(EventLog), function (error) {
                    if (error) {
                        Logger.error(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".sendEvent: EventLog save error. Error=" + error);
                    } else {
                        Logger.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".sendEvent: EventLog updated and saved. EventLog=" + JSON.stringify(EventLog));
                        cp.exec("sync", function (error, stdout, stderr) {
                            Logger.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".sendEvent: sync stdout=" + stdout);
                            if (error) {
                                Logger.error(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".sendEvent: sync error=" + error);
                            }
                        });
                    }
                });
            });
        }
        this.sendSocket(DeviceModel.port, DeviceModel.uuid, "EventLog", EventLog);// For real-time update

        //#endregion EventLog
    }
    //#endregion Class Methods (Eventing - SendEvent)

    //#region Class Methods (Eventing - Real-time Updates)
    this.sendSocket = function (port, uuid, message, parameter) {
        //Logger.warn(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".sendSocket: Port=" + port + " Message=" + message + " Parameter=" + JSON.stringify(parameter));
        SocketIo.sockets.emit(DeviceModel.id, { "Port": port, "UUID": uuid, "Message": message, "Parameter": parameter })
    }
    //#endregion Class Methods (Eventing - Real-time Updates)

    //#region Class Methods (Eventing - SMTP)
    this.sendSMTP = function (port, uuid, parameter) {

        // Send only if device's SendEventsFlag is set
        if (!SendEventsFlag) return;

        // Set up the SMTP Email message
        var message = [];
        var portNumber = (parseInt(port) + 1);
        var event = "'" + parameter.Id + "' on Port " + portNumber + " - Threshold Status=" + ThresholdValues[parameter.ThresholdState];
        message += "Timestamp=" + moment(parameter.TimeChanged).format() + "\n";
        message += "Port=" + portNumber + "\n";
        message += "Name=" + parameter.Id + "\n";
        message += "Status=" + (ThresholdValues[parameter.ThresholdState]) + "\n";
        message += "Ranged Value=" + (parameter.RangedValue) + "Volts\n";
        message += "Formula Value=" + (parameter.Formulas[parameter.FormulaIndex].FormulaValue) + "\n";
        message += "Low Threshold (Ranged)=" + (parameter.RangedLowThreshold) + "Volts\n";
        message += "Low Threshold (Formula)=" + (parameter.Formulas[parameter.FormulaIndex].FormulaLowThreshold) + "\n";
        message += "High Threshold (Ranged)=" + (parameter.RangedHighThreshold) + "Volts\n";
        message += "High Threshold (Formula)=" + (parameter.Formulas[parameter.FormulaIndex].FormulaHighThreshold) + "\n";
        message += "Hold Time=" + parseInt(parameter.HoldTime) + "\n";
        message += "Qualifier Used=" + (parameter.Qualifier ? "YES" : "NO") + "\n";

        var ioInfo = "Port: " + portNumber + " '" + parameter.Id + "'";

        Logger.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".sendEvent: Sending SMTP Event for Port " + portNumber + " '" + parameter.Id + "'");
        SmtpManager.SendSmtp(message, ioInfo, event, function (result) {
            Logger.trace(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".sendEvent: SMTP Email Event Response for Port " + portNumber + " '" + parameter.Id + "' " + result.result);
        })
    }
    //#endregion Class Methods (Eventing - SMTP)

    //#region Class Methods (Eventing - SNMP)
    this.sendSNMP = function (port, uuid, parameter) {

        // Send only if device's SendEventsFlag is set
        if (!SendEventsFlag) return;

        // Set up the SNMP trap    
        var trapOid = "1.3.6.1.4.1.11902.4." + (parseInt(port) + 1) + ".10.1.0.1";
        var varbinds = [
            {
                // Timestamp
                oid: "1.3.6.1.4.1.11902.4." + (parseInt(port) + 1) + ".1.0.1",
                type: NetSnmp.ObjectType.OctetString,
                value: new Buffer(moment().toString())
            },
            {
                // Port
                oid: "1.3.6.1.4.1.11902.4." + (parseInt(port) + 1) + ".1.0.2",
                type: NetSnmp.ObjectType.Integer,
                value: parseInt((parseInt(port) + 1))
            },
            {
                // DeviceType
                oid: "1.3.6.1.4.1.11902.4." + (parseInt(port) + 1) + ".1.0.3",
                type: NetSnmp.ObjectType.OctetString,
                value: new Buffer(DeviceModel.id)
            },
            {
                // Parameter
                oid: "1.3.6.1.4.1.11902.4." + (parseInt(port) + 1) + ".1.0.4",
                type: NetSnmp.ObjectType.OctetString,
                value: new Buffer(parameter.Id)
            },
            {
                // Value
                oid: "1.3.6.1.4.1.11902.4." + (parseInt(port) + 1) + ".1.0.5",
                type: NetSnmp.ObjectType.OctetString,
                value: new Buffer(parameter.Formulas[parameter.FormulaIndex].FormulaValue)
            },
            {
                // State
                oid: "1.3.6.1.4.1.11902.4." + (parseInt(port) + 1) + ".1.0.6",
                type: NetSnmp.ObjectType.Integer,
                value: parseInt(parameter.ThresholdState)
            },
            {
                // Units
                oid: "1.3.6.1.4.1.11902.4." + (parseInt(port) + 1) + ".1.0.7",
                type: NetSnmp.ObjectType.OctetString,
                value: new Buffer(parameter.Formulas[parameter.FormulaIndex].Units)
            }
            //,
            //{
            //    // DataString1
            //    oid: "1.3.6.1.4.1.11902.4." + (parseInt(port) + 1) + ".1.0.10",
            //    type: NetSnmp.ObjectType.OctetString,
            //    value: new Buffer(" ")
            //},
            //{
            //    // DataString2
            //    oid: "1.3.6.1.4.1.11902.4." + (parseInt(port) + 1) + ".1.0.11",
            //    type: NetSnmp.ObjectType.OctetString,
            //    value: new Buffer(" ")
            //},
            //{
            //    // DataString3
            //    oid: "1.3.6.1.4.1.11902.4." + (parseInt(port) + 1) + ".1.0.12",
            //    type: NetSnmp.ObjectType.OctetString,
            //    value: new Buffer(" ")
            //},
            //{
            //    // DataString4
            //    oid: "1.3.6.1.4.1.11902.4." + port + ".1.0.13",
            //    type: NetSnmp.ObjectType.OctetString,
            //    value: new Buffer(" ")
            //},
            //{
            //    // DataString5
            //    oid: "1.3.6.1.4.1.11902.4." + port + ".1.0.14",
            //    type: NetSnmp.ObjectType.OctetString,
            //    value: new Buffer(" ")
            //},
            //{
            //    // DataInteger1
            //    oid: "1.3.6.1.4.1.11902.4." + port + ".1.0.20",
            //    type: NetSnmp.ObjectType.Integer,
            //    value: parseInt(0)
            //},
            //{
            //    // DataInteger2
            //    oid: "1.3.6.1.4.1.11902.4." + port + ".1.0.21",
            //    type: NetSnmp.ObjectType.Integer,
            //    value: parseInt(0)
            //},
            //{
            //    // DataInteger3
            //    oid: "1.3.6.1.4.1.11902.4." + port + ".1.0.22",
            //    type: NetSnmp.ObjectType.Integer,
            //    value: parseInt(0)
            //},
            //{
            //    // DataInteger4
            //    oid: "1.3.6.1.4.1.11902.4." + port + ".1.0.23",
            //    type: NetSnmp.ObjectType.Integer,
            //    value: parseInt(0)
            //},
            //{
            //    // DataInteger5
            //    oid: "1.3.6.1.4.1.11902.4." + port + ".1.0.24",
            //    type: NetSnmp.ObjectType.Integer,
            //    value: parseInt(0)
            //}
        ];

        // Send the SNMP notification
        for (var i = 0; i < SnmpHostSessionsArray.length; i++) {
            // Skip if this host is not enabled
            if (SnmpHostSessionsArray[i].Enabled === "FALSE") continue;

            // Send trap to host ...
            var IPAddress = SnmpHostSessionsArray[i].Session.target;
            Logger.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".sendSNMP: Sending SNMP trap to IPAddress=" + IPAddress);

            SnmpHostSessionsArray[i].Session.trap(trapOid, varbinds, function (error) {
                if (error) {
                    Logger.error(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".sendSNMP: Error=" + error);
                }
                else {
                    Logger.trace(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".sendSNMP: Sending to SNMP Host @ " + IPAddress + " - Port[" + varbinds[1].value + "] '" + varbinds[2].value + "' Value=" + varbinds[4].value);
                }
            });
        }

    }
    //#endregion Class Methods (Eventing - SNMP)

    //#endregion Class Methods (Eventing)

    //#endregion Class Methods
}

//#region Device Constructor
// Device constructor to ensure unique objects
//  Uses the following signature:
//      port        - port on which device is installed [0 indexed]
//      logger      - system logger object
//      io          - system socket io object
//      smtpManager - SMTP messaging object
//      netSnmp     - SNMP infrastructure object
module.exports = function (port, logger, io, smtpManager, netSnmp, callback) {
    var DeviceObject = new Device(port, logger, io, smtpManager, netSnmp);
    callback(DeviceObject);
}
//#endregion Device Constructor
//#endregion Device Class: Variables, Instantiation, Methods and Constructor

// Exported Methods
//#region Exported Methods

// Basic Info
//#region Basic Info
Device.prototype.SetPort = function (port) {
    this.setPort(port);
}

Device.prototype.GetPort = function () {
    return this.getPort()
}

Device.prototype.GetDeviceId = function () {
    return this.getDeviceId()
}

Device.prototype.GetDeviceModel = function () {
    return this.getDeviceModel()
}

Device.prototype.GetDeviceInformation = function (callback) {
    var deviceInfo = this.getDeviceInformation();
    callback(deviceInfo);
}

//#endregion Basic Info

// Enable/Disable
//#region Enable/Device
Device.prototype.Enable = function (enabled) {
    this.enable(enabled);
}
//#endregion Enable/Disable

// Send Events
//#region Send Events
Device.prototype.SendEvents = function (sendevents) {
    var id = this.getDeviceId();
    Logger.trace(id + ".SendEvents: SendEventsFlag=" + (sendevents ? "YES" : "NO"));
    this.sendEvents(sendevents);
}
//#endregion Send Events

// Update Device
//#region Update Device

// UpdateDevice
Device.prototype.UpdateDevice = function (port, model, callback) {
    var id = this.getDeviceId();
    this.updateDevice(port, model, function (model) {
        Logger.trace(id + ".UpdateDevice: Updated DeviceModel=" + JSON.stringify(model));
        callback(model);
    });
}

// SetSnmpHostSessionsArray
Device.prototype.SetSnmpHostSessionsArray = function (snmpHostSessions, callback) {
    var id = this.getDeviceId();
    SnmpHostSessionsArray = snmpHostSessions.slice(0);// Copy array
    Logger.trace(id + ".SetSnmpHostSessionsArray: SNMP Host Sessions established");
    callback();
}

//#endregion Update Device

//#endregion ExportedMethods